package ps2;

import java.util.*;
import java.io.*;

import techniques.BN.*;
import techniques.BN.solver.*;

/**
 *  Implements the variable elimination algorithm for calcluating
 *  conditional probability on Bayesian networks.
 */
public class VariableEliminationSolver extends Solver {

    public static final boolean DEBUG = false;

    protected boolean random = false;

    public VariableEliminationSolver(BayesNet _bn) {
	super(_bn);
    }

    public void setRandom(boolean random) {
	this.random = random;
    }

    public double[] getProb(Variable variable) {
	if (random) {
	    List nodes = this.bn.getNodesWithTopologicalOrdering();
	    for (int i = 0; i < nodes.size(); i++) {
		int toSwapIndex = techniques.utils.Random.random(nodes.size()-i) + i;
		Variable toSwap = (Variable)nodes.get(toSwapIndex);
		nodes.set(toSwapIndex, nodes.get(i));
		nodes.set(i, toSwap);
	    }
	    System.out.println("Using elimination order: " + nodes);
	    return this.getProb(variable, nodes);
	}
	else {
	    ArrayList reversed = new ArrayList();
	    List topo = this.bn.getNodesWithTopologicalOrdering();
	    for (int i = topo.size()-1; i >= 0; i--) {
		reversed.add(topo.get(i));
	    }
	    
	    return this.getProb(variable, reversed);
	}
    }


    /**
     *  Performs variable elimination on this Bayes Net using the
     *  specified elimination order.
     *
     *  @param elimOrder a list of Variable objects in the order they
     *  should be eliminated.
     */
    public double[] getProb(Variable variable, List elimOrder) {
	HashSet factors = buildInitialFactors();

	for (int i = 0; i < elimOrder.size(); i++) {
	    if (DEBUG) System.out.println("Eliminating " + elimOrder.get(i));
	    if (DEBUG) System.out.println("  Current factors: " + factors);

	    Variable currVar = (Variable)elimOrder.get(i);
	    if (!currVar.equals(variable) && // ???
		!this.evidence.containsKey(currVar)) {	    // current var is not in evidence, so we have to sum.

		// pull out factors that depend on currVar
		HashSet dependentFactors = new HashSet();
		Factor[] f = (Factor[])factors.toArray(new Factor[0]);
		for (int j = 0; j < f.length; j++) {
		    if (f[j].dependsOn(currVar)) {
			factors.remove(f[j]);
			dependentFactors.add(f[j]);
		    }
		}

		if (DEBUG) System.out.println("==> " + currVar + " not in evidence, summing out over factors " + dependentFactors + ":");

		Factor summedOut = sumOutVar(currVar, dependentFactors);
		factors.add(summedOut);
	    }
	}

	double[] probs = calculateProbsGivenVariable(variable, factors);
	
	return ProbUtils.normalize(probs);
    }

    /**
     *  Calculates the probabilities of the given variable over the
     *  given factors.
     */
    public double[] calculateProbsGivenVariable(Variable variable, HashSet factors) {
	double[] probs = new double[variable.getNumValues()];
	
	for (int j = 0; j < variable.getNumValues(); j++) {
	    HashMap currEvidence = (HashMap)this.evidence.clone();
	    currEvidence.put(variable, new Integer(j));
	    
	    Iterator factorIter = factors.iterator();
	    double prob = ((Factor)factorIter.next()).getProb(currEvidence);
	    while (factorIter.hasNext()) {
		prob *= ((Factor)factorIter.next()).getProb(currEvidence);
	    }
	    probs[j] = prob;
	}

	return probs;
    }

    // build initial factors (just CPTs from given BN)	
    public HashSet buildInitialFactors() {
	HashSet factors = new HashSet();
	for (int i = 0; i < this.bn.getNodes().size(); i++) {
	    Variable currVar = (Variable)this.bn.getNodes().get(i);
	    Variable[] currParents  = currVar.getParents();

	    ArrayList dependentVarsList = new ArrayList();
	    if (!this.evidence.containsKey(currVar))
		dependentVarsList.add(currVar);
	    for (int j = 0; j < currParents.length; j++) {
		if (!this.evidence.containsKey(currParents[j]))
		    dependentVarsList.add(currParents[j]);
	    }

	    Factor currFactor = new Factor(currVar,
					   (Variable[])dependentVarsList.toArray(new Variable[0]),
					   (HashMap)this.evidence.clone());
	    if (DEBUG) System.out.println("Created Factor: " + currFactor);
	    factors.add(currFactor);
	}
	return factors;
    }


    /**
     *  Combines the given Factors, using the pointwise product and
     *  summing out the given Variable over its possible values.
     * 
     *  @param summationVar the variable being summed out
     *  @param factors the factors to be summed
     *  @return a new Factor with the combined values from the input Factors
     */
    public Factor sumOutVar(Variable summationVar, HashSet factors) {
	HashSet newFactorVars = new HashSet();
	HashSet newDependentVars = new HashSet();

	Iterator factorIter = factors.iterator();
	while (factorIter.hasNext()) {
	    Factor currFactor = (Factor)factorIter.next();
	    for (int j = 0; j < currFactor.factorVars.length; j++) 
		newFactorVars.add(currFactor.factorVars[j]);
	    for (int j = 0; j < currFactor.dependentVars.length; j++) 
		newDependentVars.add(currFactor.dependentVars[j]);
	}
	
	newDependentVars.remove(summationVar);

	Factor f = new Factor((Variable[])newFactorVars.toArray(new Variable[0]),
			      (Variable[])newDependentVars.toArray(new Variable[0]));

	if (DEBUG) System.out.println("    Created factor " + f);
	setPointwiseVals(f, factors, new HashSet(newDependentVars), this.evidence, summationVar);

	return f;
    }

    protected void setPointwiseVals(Factor f, HashSet factors, HashSet remainingVars, HashMap currEvidence, Variable summationVar) {
	if (remainingVars.isEmpty()) {
	    //	    if (DEBUG) System.out.println("    No remaining vars, summing over " + summationVar + " with evidence " + currEvidence);
	    double sum = 0;
	    for (int i = 0; i < summationVar.getNumValues(); i++) {
		//		if (DEBUG) System.out.println("      sum = " + sum);
		HashMap ev = (HashMap)currEvidence.clone();
		ev.put(summationVar, new Integer(i));
		
		Iterator factorIter = factors.iterator();
		double prob = ((Factor)factorIter.next()).getProb(ev);
		//		if (DEBUG) System.out.println("      prob = " + prob);
		while (factorIter.hasNext()) {
		    //		    if (DEBUG) System.out.println("      prob = " + prob);
		    prob *= ((Factor)factorIter.next()).getProb(ev);
		}

		sum += prob;
	    }

	    f.setProb(currEvidence, sum);
	}
	else {
	    Variable next = (Variable)remainingVars.iterator().next();
	    HashSet nextRemainingVars = (HashSet)remainingVars.clone();
	    nextRemainingVars.remove(next);
	    for (int i = 0; i < next.getNumValues(); i++) {
		HashMap nextEvidence = (HashMap)currEvidence.clone();
		nextEvidence.put(next, new Integer(i));
		setPointwiseVals(f, factors, nextRemainingVars, nextEvidence, summationVar);
	    }
	}
    }


    public static void main(String[] args) throws Exception {
	try {
	    BufferedWriter out = null;

 	    BayesNet burg = techniques.BN.Nets.getBurglary();
 	    Solver burgEngine = new VariableEliminationSolver(burg);

   	    BayesNet ins = techniques.BN.Nets.getInsurance();
   	    VariableEliminationSolver insEngine = new VariableEliminationSolver(ins);

 	    BayesNet carpo = techniques.BN.Nets.getCarpo();
 	    VariableEliminationSolver carpoEngine = new VariableEliminationSolver(carpo);

 	    // 1.1	    
  	    burgEngine.setEvidenceGivenVal(Evidence.get("1.1"));
  	    ProbUtils.printProbs(burgEngine, burg.getNode("Burglary"));

  	    // 1.2 
  	    insEngine.setEvidenceGivenVal(Evidence.get("1.2"));
  	    ProbUtils.printProbs(insEngine, ins.getNode("PropCost"));

 	    // 2.a 
 	    insEngine.setEvidenceGivenVal(Evidence.get("2.a.1"));
 	    ProbUtils.printProbs(insEngine, ins.getNode("PropCost"));
 	    insEngine.setEvidenceGivenVal(Evidence.get("2.a.2"));
 	    ProbUtils.printProbs(insEngine, ins.getNode("PropCost"));

 	    // 2.b
 	    carpoEngine.setEvidenceGivenVal(Evidence.get("2.b.1"));
 	    ProbUtils.printProbs(carpoEngine, carpo.getNode("N112"));
 	    carpoEngine.setEvidenceGivenVal(Evidence.get("2.b.2"));
 	    ProbUtils.printProbs(carpoEngine, carpo.getNode("N143"));

	    // 3.a
// 	    insEngine.setEvidenceGivenVal(Evidence.get("2.a.1"));
// 	    insEngine.setRandom(true);
// 	    out = new BufferedWriter(new FileWriter("ps2/prob3.a.1.dat"));
// 	    for (int i = 0; i < 50; i++) {
// 	        System.out.println(">> loop " + i);
// 		try {
// 		    long time = ProbUtils.printProbs(insEngine, ins.getNode("PropCost"));
// 		    out.write(time+"\n");
// 		}
// 		catch (Throwable e) {
// 		    System.out.println(e);
// 		    out.write("exception\n");
// 		}
// 		out.flush();
// 	    }
// 	    out.close();

// 	    insEngine.setEvidenceGivenVal(Evidence.get("2.a.2"));
// 	    insEngine.setRandom(true);
// 	    out = new BufferedWriter(new FileWriter("ps2/prob3.a.2.dat"));
// 	    for (int i = 0; i < 50; i++) {
// 	        System.out.println(">> loop " + i);
// 		try {
// 		    long time = ProbUtils.printProbs(insEngine, ins.getNode("PropCost"));
// 		    out.write(time+"\n");
// 		}
// 		catch (Throwable e) {
// 		    System.out.println(e);
// 		    out.write("exception\n");
// 		}
// 		out.flush();
// 	    }
// 	    out.close();

//   	    // 3.b
//  	    carpoEngine.setEvidenceGivenVal(Evidence.get("2.b.1"));
//  	    carpoEngine.setRandom(true);
//  	    out = new BufferedWriter(new FileWriter("ps2/prob3.b.1.dat"));
//  	    for (int i = 0; i < 50; i++) {
//  	        System.out.println(">> loop " + i);
//  		try {
//  		    long time = ProbUtils.printProbs(carpoEngine, carpo.getNode("N112"));
//  		    out.write(time+"\n");
//  		}
//  		catch (Throwable e) {
//  		    System.out.println(e);
//  		    out.write("exception\n");
//  		}
//  		out.flush();
//  	    }
//  	    out.close();

//  	    carpoEngine.setEvidenceGivenVal(Evidence.get("2.b.2"));
//  	    carpoEngine.setRandom(true);
//  	    out = new BufferedWriter(new FileWriter("ps2/prob3.b.2.dat"));
//  	    for (int i = 0; i < 50; i++) {
//  	        System.out.println(">> loop " + i);
//  		try {
//  		    long time = ProbUtils.printProbs(carpoEngine, carpo.getNode("N143"));
//  		    out.write(time+"\n");
//  		}
//  		catch (Throwable e) {
//  		    System.out.println(e);
//  		    out.write("exception\n");
//  		}
//  		out.flush();
//  	    }
//  	    out.close();

	}
	catch (Exception e) {
	    e.printStackTrace(System.out);
	}
    }


}


/**
 *  Represents a factor of several variables - essentially a wrapper
 *  for ProbTable.
 */
class Factor {

    public static boolean DEBUG;

    static {
	DEBUG = VariableEliminationSolver.DEBUG;
    }

    protected ProbTable cpt;
    protected Variable[] dependentVars;
    protected Variable[] factorVars;
    protected HashMap evidence;

    /**
     *  Creates a factor of one variable.
     *  @param factorVar the variable in the subscript
     *  @param dependentVars all variables of which this factor is a function 
     *  @param evidence current evidence with which to create this factor
     */
    public Factor(Variable factorVar, Variable[] dependentVars, HashMap evidence) {
	this.cpt = new ProbTable(dependentVars);
	this.factorVars = new Variable[] {factorVar};
	this.dependentVars = dependentVars;
	this.evidence = evidence;

	this.cpt = factorVar.getProbTable();
	/*	HashSet remainingVars = new HashSet();
	for (int i = 0; i < dependentVars.length; i++) remainingVars.add(dependentVars[i]);
	this.setProbs(factorVar, remainingVars, evidence);*/
    }

    protected Factor(Variable[] factorVars, Variable[] dependentVars) {
	this.cpt = new ProbTable(dependentVars);
	this.dependentVars = dependentVars;
	this.factorVars = factorVars;
	this.evidence = new HashMap();
    }

    /*    protected void setProbs(Variable factorVar, HashSet remainingVars, HashMap currEvidence) {
	if (remainingVars.size() == 1) {
	    Variable lastVar = (Variable)remainingVars.iterator().next();
	    for (int i = 0; i < lastVar.getNumValues(); i++) {
		HashMap lastEvidence = (HashMap)currEvidence.clone();
		lastEvidence.put(lastVar, new Integer(i));
		if (DEBUG) System.out.println("  Setting probs for " + this + " from var " + factorVar + " with evidence " + lastEvidence + " to " + factorVar.getProb(i, lastEvidence));
		this.cpt.setProb(lastEvidence, factorVar.getProb(i, lastEvidence));
	    }
	}
	else {
	    Variable nextVar = (Variable)remainingVars.iterator().next();
	    HashSet nextRemaining = (HashSet)remainingVars.clone();
	    nextRemaining.remove(nextVar);
	    for (int i = 0; i < nextVar.getNumValues(); i++) {
		HashMap nextEvidence = (HashMap)currEvidence.clone();
		nextEvidence.put(nextVar, new Integer(i));
		setProbs(factorVar, nextRemaining, nextEvidence);
	    }
	}
	}*/

    public String toString() {
	StringBuffer out = new StringBuffer();
	out.append("F_");
	for (int i = 0; i < factorVars.length; i++) {
	    if (factorVars[i].getName().length() > 0) {
		out.append(factorVars[i].getName().substring(0,1));
	    }
	}
	out.append("(");
	for (int i = 0; i < dependentVars.length; i++) {
	    if (dependentVars[i].getName().length() > 0) {
		out.append(dependentVars[i].getName().substring(0,1));
	    }
	    if (i != dependentVars.length-1) {
		out.append(",");
	    }
	}
	out.append(")");
	return out.toString();
    }

    /**
     *  Two Factors are equal if their factorVars contain the same Variables.
     */
    public boolean equals(Object other) {
	Variable[] otherFactorVars = ((Factor)other).factorVars;
	if (this.factorVars.length != otherFactorVars.length) return false;

	Arrays.sort(this.factorVars, this.getVariableComparator());
	Arrays.sort(otherFactorVars, this.getVariableComparator());
	for (int i = 0; i < this.factorVars.length; i++) {
	    if (!this.factorVars[i].getName().equals(otherFactorVars[i].getName())) return false;
	}
	return true;
    }

    protected Comparator getVariableComparator() {
	return new Comparator() {
		public int compare(Object o1, Object o2) {
		    return ((Variable)o1).getName().compareTo(((Variable)o2).getName());
		}
		public boolean equals(Object other) {
		    return this == other;
		}
	    };
    }

    /**
     *  Returns true if this Factor depends on the given Variable
     */
    public boolean dependsOn(Variable v) {
	for (int i = 0; i < this.dependentVars.length; i++) {
	    if (this.dependentVars[i].equals(v)) return true;
	}
	return false;
    }

    /**
     *  @param values maps Variables to Integers 
     */
    public double getProb(Map values) {
	return cpt.getProb(values);
    }

    public void setProb(Map evidence, double prob) {
	this.cpt.setProb(evidence, prob);
    }


}


/*
 *  want to eliminate one variable at a time.
 *  means finding all factors with that variable mentioned
 *  means finding P(var|parents) and sum{ P(children(var)|parents) }
 *
 */
