/*
 * Created on Oct 9, 2003
 *
 */
package techniques.BN;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.NoSuchElementException;

/**
 * Represents a probability table, and can be thought of as a maintainer of 
 * functions of discrete types.
 * 
 * @author vineet
 */
public class ProbTable {
	protected Variable[] variables = null;
	public double[] table = null;

	public ProbTable(Variable var) {
		variables = new Variable[] { var };
		table = new double[getTableSize()];

		// let's initialize to -1 to help find errors
		for (int i = 0; i < table.length; i++) {
			table[i] = -1;
		}
	}

	public ProbTable(Variable[] var) {
		variables = var;
		int tblSize = getTableSize();
		if (tblSize > 1000000) {
			//System.err.print("<<largeProbTable: " + tblSize + ">>");
			System.err.println(
				"Constructing ProbTable: " +				"Allocating large probability table of size: "
					+ tblSize
					+ " for: "
					+ getVarString());
		}
		table = new double[tblSize];

		// let's initialize to -1 to help find errors
		for (int i = 0; i < tblSize; i++) {
			table[i] = -1;
		}
	}

	/**
	 * Returns size of the underlying table.
	 */
	public int getTableSize() {
		int retVal = 1;
		for (int i = 0; i < variables.length; i++) {
			retVal = retVal * variables[i].getNumValues();
		}
		return retVal;
	}

	/**
	 * Returns size of the table that would be formed from the given variables
	 */
	public static int getTableSize(Variable[] var) {
		int retVal = 1;
		for (int i = 0; i < var.length; i++) {
			retVal = retVal * var[i].getNumValues();
		}
		return retVal;
	}

	/**
	 * Sets the probability table value given all the values for the parents
	 * of the current prob. table. This method is identical to setProb 
	 */
	public void setTableValue(HashMap knownValues, double inTableValue) {
		setTableValues(knownValues, new double[] { inTableValue });
	}

	/**
	 * Sets the probability table value given all the values for the parents
	 * of the current prob. table.
	 */
	public void setTableValue(int[] evidence, double inTableValue) {
		table[getEvidenceIndex(evidence)] = inTableValue;
	}

	/**
	 * Sets the probability tables values given the first few variable values
	 * for the current prob. table. 
	 */
	public void setTableValues(HashMap knownValues, double[] inTableValues) {
		int dataNdx = 0;
		int dataLen = 1;
		for (int i = 0; i < variables.length; i++) {
			Integer val = (Integer) knownValues.get(variables[i]);
			if (val != null && dataLen == 1) {
				dataNdx =
					dataNdx * variables[i].getNumValues() + val.intValue();
			} else {
				dataLen = dataLen * variables[i].getNumValues();
				dataNdx = dataNdx * variables[i].getNumValues();
			}
		}

		if (inTableValues.length != dataLen) {
			throw new RuntimeException(
				"Expected length of inTableValues: "
					+ dataLen
					+ " [got:"
					+ inTableValues.length
					+ "]");
		}

		for (int i = 0; i < inTableValues.length; i++) {
			table[dataNdx + i] = inTableValues[i];
		}
	}

	/**
	 * Expects complete evidence
	 */
	protected int getEvidenceIndex(Map evidence) {
		int dataNdx = 0;
		for (int i = 0; i < variables.length; i++) {
			Integer parValNdx = (Integer) evidence.get(variables[i]);
			if (parValNdx == null) {
				throw new RuntimeException(
					"Not enough evidence, expected: " + variables[i]);
			}
			dataNdx =
				(dataNdx * variables[i].getNumValues()) + parValNdx.intValue();
		}
		return dataNdx;
	}

	/**
	 * Expects complete evidence
	 */
	protected int getEvidenceIndex(int[] evidence) {
		int dataNdx = 0;
		for (int i = 0; i < variables.length; i++) {
			dataNdx = (dataNdx * variables[i].getNumValues()) + evidence[i];
		}
		return dataNdx;
	}

	/**
	 * Used to get the probability values for the given variable evidence. All
	 * values (as given by indices) in the table (for the variable and parents), 
	 * need to be specified in the evidence (otherwise a not enough evidence 
	 * exception will be thrown). 
	 */
	public double getProb(Map evidence) {
		return table[getEvidenceIndex(evidence)];
	}

	/**
	 * All values for variables in the table are needed to be in the map for 
	 * successful completion of the method 
	 */
	public void setProb(Map evidence, double prob) {
		table[getEvidenceIndex(evidence)] = prob;
	}

	/**
	 * Same as getProb(Map) with evidence instead provided as an int array.
	 * @param evidence -  evidence in the same order as the variables
	 */
	public double getProb(int[] evidence) {
		return table[getEvidenceIndex(evidence)];
	}

	/**
	 * @param evidence -  evidence in the same order as the variables
	 */
	public void setProb(int[] evidence, double prob) {
		table[getEvidenceIndex(evidence)] = prob;
	}

	public Variable[] getVariables() {
		return variables;
	}
	public Variable getVariable(int i) {
		return variables[i];
	}
	public int getNumVariables() {
		return variables.length;
	}
	/**
	 * Returns string of variables in table
	 */
	public String getVarString() {
		StringBuffer sb = new StringBuffer();
		sb.append("<");
		for (int i = 0; i < variables.length; i++) {
			if (i != 0) {
				sb.append(",");
			}
			sb.append(variables[i]);
		}
		sb.append(">");
		return sb.toString();
	}
	public String toString() {
		return getVarString();
	}
	/**
	 * Returns string of variables and data in table
	 */
	public String getCompleteString() {
		StringBuffer sb = new StringBuffer();
		sb.append(getVarString());
		sb.append(" {");
		for (int i = 0; i < table.length; i++) {
			if (i != 0) {
				sb.append(",");
			}
			sb.append(table[i]);
		}
		sb.append("}");
		return sb.toString();
	}

	/**
	 * Class works by storing the next entry at each step, which is updated
	 * when next is called
	 * 
	 * @author vineet
	 */
	private class VariableEntriesIterator implements Iterator {
		HashMap entry = null;

		public VariableEntriesIterator() {
			entry = new HashMap();
			for (int i = 0; i < variables.length; i++) {
				entry.put(variables[i], new Integer(0));
			}
		}

		/* (non-Javadoc)
		 * @see java.util.Iterator#remove()
		 */
		public void remove() {
			throw new UnsupportedOperationException();
		}

		/* (non-Javadoc)
		 * @see java.util.Iterator#hasNext()
		 */
		public boolean hasNext() {
			if (entry != null) {
				return true;
			} else {
				return false;
			}
		}

		/* (non-Javadoc)
		 * @see java.util.Iterator#next()
		 */
		public Object next() {
			HashMap nextEntry = (HashMap) entry.clone();

			// update entry! [by one]
			if (entry == null) {
				throw new NoSuchElementException();
			}
			for (int i = 0; i < variables.length; i++) {
				int entryVarValue =
					((Integer) entry.get(variables[i])).intValue();
				if (entryVarValue + 1 < variables[i].getNumValues()) {
					entry.put(variables[i], new Integer(entryVarValue + 1));
					//continue makeFactorIncrEntry;
					return nextEntry;
				} else {
					entry.put(variables[i], new Integer(0));
				}
			}

			// can't incr.
			entry = null;
			return nextEntry;
		}
	}

	/**
	 * Returns an iterator through a set of all entries (variable-value 
	 * combinations for the variables) in the table. The entries are represented 
	 * by maps of the variables to their values (actually indexes of the values).
	 */
	public Iterator getVariablesIterator() {
		return new VariableEntriesIterator();
	}

	/*
	public void testTable() {
		// test for -1
		int negCnt = 0;
		Iterator it = this.getVariablesIterator();
		while (it.hasNext()) {
			HashMap hm = (HashMap) it.next();
			if (this.getProb(hm) < 0) {
				negCnt++;
				//System.err.println("Value less than -1");
			}
		}
			System.err.println(this +" negative counts: " + negCnt);
		// test for sum to 1
	}
	*/
}
