/*
 * Created on Oct 6, 2003
 *
 */
package techniques.BN;

import java.util.HashMap;

/**
 * Represents a variable in a bayes net, and beyond the variable name stores 
 * the possible values and the conditional probability table for the variable.
 * By default variables have 2 <i>given</i> values (1) Boolean.FALSE, and 
 * (2) Boolean.TRUE . These values are mapped to numbers (from 0 to getNumValues)
 * and programs need to explicitly retrieve the <i>given</i> values for the non
 * numeric values.
 *   
 * @author vineet
 */
public class Variable implements Cloneable {
	protected String name = null;
	protected Object[] values = null;
	protected Variable[] parents = null;
	protected ProbTable cpt = null;

	/**
	 * @param varName - name of variable
	 */
	public Variable(String varName) {
		name = varName;
		values = new Object[] { Boolean.FALSE, Boolean.TRUE };
		parents = new Variable[] {
		};
		cpt = new ProbTable(this);
	}

	/**
	 * @param varName - name of variable
	 * @param varGivenValues - given values of the variables (note all variables
	 * have values which are integers, and given values, for example Boolean.TRUE
	 * and Boolean.FALSE
	 */
	public Variable(String varName, Object[] varGivenValues) {
		name = varName;
		values = varGivenValues;
		parents = new Variable[] {
		};
		cpt = new ProbTable(this);
	}

	/**
	 * Returns variable name
	 */
	public String getName() {
		return name;
	}

	/**
	 * Returns cond. prob. table
	 */
	public ProbTable getProbTable() {
		return cpt;
	}

	public String toString() {
		return name;
	}

	/**
	 * Returns all given values (in an array with indices representing the 
	 * mapping from number to given value)
	 */
	public Object[] getGivenValues() {
		return values;
	}

	/**
	 * Just returns all numbers from 0 to getNumValues()
	 */
	public int[] getValues() {
		int[] retVal = new int[values.length];
		for (int i = 0; i < values.length; i++) {
			retVal[i] = i;
		}
		return retVal;
	}

	/**
	 * Maps given value to number
	 */
	public int getValNdx(Object givenValue) {
		for (int i = 0; i < values.length; i++) {
			if (values[i].equals(givenValue)) {
				return i;
			}
		}
		throw new RuntimeException(
			"Value: " + givenValue + " not in variable: " + name);
	}

	/**
	 * Maps number to given value
	 */
	public Object getGivenValue(int ndx) {
		return values[ndx];
	}


	/**
	 * Returns number of values used by variable
	 */
	public int getNumValues() {
		return values.length;
	}

	/**
	 * set parents for the conditional prob. table
	 */
	public void setParent(Variable var) {
		setParents(new Variable[] { var });
	}

	/**
	 * set parents for the conditional prob. table
	 */
	public void setParents(Variable[] vars) {
		parents = vars;

		Variable[] cptVars = new Variable[vars.length + 1];
		// +1 for current variable
		cptVars[0] = this;
		for (int i = 0; i < vars.length; i++) {
			cptVars[i + 1] = vars[i];
		}
		cpt = new ProbTable(cptVars);
	}

	public Variable[] getParents() {
		return parents;
	}

	/**
	 * Used to set the condition probability table values by only providing
	 * values for variable given value being Boolean.TRUE (the method calculate
	 * values for Boolean.FALSE).
	 */
	public void setTrueProb(double[] trueCPT) {
		setProb(getValNdx(Boolean.TRUE), trueCPT);
		double[] falseCPT = new double[trueCPT.length];
		for (int i = 0; i < trueCPT.length; i++) {
			falseCPT[i] = 1 - trueCPT[i];
		}
		setProb(getValNdx(Boolean.FALSE), falseCPT);
	}

	/**
	 * Used to set the condition probability table values for the given variable
	 * vaue index/number
	 */
	public void setProb(int i, double[] valueCPT) {
		HashMap varKnown = new HashMap();
		varKnown.put(this, new Integer(i));
		cpt.setTableValues(varKnown, valueCPT);
	}

	/**
	 * Used to get the probability values for the given variable
	 * vaue index/number and evidence
	 * @param value - variable value index/number
	 * @param evidence - a map of variables to their value indices 
	 */
	public double getProb(int value, HashMap evidence) {
		HashMap probTableEvidence = (HashMap) evidence.clone();
		probTableEvidence.put(this, new Integer(value));
		return cpt.getProb(probTableEvidence);
	}

	/**
	 * Used to get the probability values for the given variable evidence. All
	 * values (as given by indices) in the table (for the variable and parents), 
	 * need to be specified in the evidence (otherwise a not enough evidence 
	 * exception will be thrown). 
	 */
	public double getProb(HashMap evidence) {
		return cpt.getProb(evidence);
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#clone()
	 */
	public Object clone() throws CloneNotSupportedException {
		return super.clone();
	}


    /**
     *  MODIFICATION: ahogue - added equals() and hashCode()
     */
    public boolean equals(Object other) {
	return this.getName().equals(((Variable)other).getName());
    }
    public int hashCode() {
	return this.getName().hashCode();
    }
    

}
