/*
 * Created on Oct 7, 2003
 *
 */
package techniques.BN.solver;

import java.util.HashMap;
import java.util.LinkedList;

import techniques.BN.*;

/**
 * An implementation of the most basic solver for the BayesNets. The algorithm
 * has been implemented as per the pseudo-code provided in 
 * AIMA, 2nd Ed., pg. 506
 * 
 * @author vineet
 */
public class EnumerationSolver extends Solver {

    public static final boolean DEBUG = false;

    /**
     * @param _bn
     */
    public EnumerationSolver(BayesNet _bn) {
	super(_bn);
    }

    private double[] enumerationAsk(Variable var) {
	double[] retVal = new double[var.getNumValues()];
	double sum = 0;
	for (int value = 0; value < var.getNumValues(); value++) {
	    // copy evidence and set var to one of its possible values
	    HashMap valEvidence = (HashMap) evidence.clone();
	    valEvidence.put(var, new Integer(value));
	    if (DEBUG) System.out.println(var + "=" + value);

	    retVal[value] =
		enumerateAll(
			     "  ",
			     bn.getNodesWithTopologicalOrdering(),
			     valEvidence);
	    sum += retVal[value];
	}
	for (int i = 0; i < var.getNumValues(); i++) {
	    retVal[i] = retVal[i] / sum;
	}
	return retVal;
    }

    // finds probability of currEvidence (sums over all vars _not_ in evidence)
    private double enumerateAll(
				String h,
				LinkedList vars,
				HashMap currEvidence) {
	if (vars.isEmpty()) {
	    return 1.0;
	}
	Variable Y = (Variable) vars.iterator().next();
	LinkedList restVars = (LinkedList) vars.clone();
	restVars.remove(Y);
	Integer evidenceValueY = (Integer) currEvidence.get(Y);

	// is current var (Y) in evidence?
	if (evidenceValueY != null) { // YES
	    if (DEBUG) System.out.println(
			       h
			       + "P("
			       + Y
			       + "="
			       + Y.getGivenValue(((Integer) evidenceValueY).intValue())
			       + ")");
	    // Y is fixed, so multiply its prob by the prob of the rest of the vars
	    return Y.getProb(evidenceValueY.intValue(), currEvidence)
		* enumerateAll(h + "  ", restVars, currEvidence);
	}
	else {			// NO
	    double sum = 0;
	    if (DEBUG) System.out.println(h + "E [" + Y + "]");
	    // Y is not fixed, so sum over all its possible values
	    for (int value = 0; value < Y.getNumValues(); value++) {
		HashMap valEvidence = (HashMap) currEvidence.clone();
		valEvidence.put(Y, new Integer(value));
		//System.out.println(h+"P(" + Y + "=" + Y.getGivenValue(value) + ")");
		if (DEBUG) System.out.println(h + "|-" + Y.getGivenValue(value));

		// recurse, adding prob of curr value of Y times prob of rest of evidence
		sum += Y.getProb(value, currEvidence)
		    * enumerateAll(h + "  ", restVars, valEvidence);
	    }
	    return sum;
	}
    }

    /**
     * @param variable
     */
    public double[] getProb(Variable variable) {
	return enumerationAsk(variable);
    }

    public static void main(String[] args) throws Exception {
	try {
	    /* 1.1
	    BayesNet bn = Nets.getBurglary();
	    
	    Solver engine = new EnumerationSolver(bn);
	    HashMap e = new HashMap();
	    e.put(bn.getNode("JohnCalls"), Boolean.TRUE);
	    e.put(bn.getNode("MaryCalls"), Boolean.TRUE);
	    engine.setEvidenceGivenVal(e);
	    
	    Variable burgVar = bn.getNode("Burglary");
	    System.out.println("Probabilitiy: " + engine.getProb(burgVar)[burgVar.getValNdx(Boolean.TRUE)]);
	    */
	    
	    /* 1.2 */
	    BayesNet bn = techniques.BN.Nets.getInsurance();
	    
	    Solver engine = new EnumerationSolver(bn);
	    HashMap e = new HashMap();
	    e.put(bn.getNode("Age"), "Adolescent");
	    e.put(bn.getNode("Antilock"), "False");
	    e.put(bn.getNode("Mileage"), "FiftyThou");
	    engine.setEvidenceGivenVal(e);

	    Variable propCostVar = bn.getNode("PropCost");
	    double[] propCostProbs = engine.getProb(propCostVar);
	    System.out.println("Probabilities: ");
	    for (int i = 0; i < propCostVar.getNumValues(); i++) {
		System.out.println("  " + propCostVar.getGivenValue(i) + " = " + propCostProbs[i]);
	    }
	}
	catch (Exception e) {
	    e.printStackTrace(System.out);
	}
    }


}
